# PYTHON_ARGCOMPLETE_OK
import argparse
import os
import subprocess
import sys
from time import time, strftime, gmtime

from .batch_runner import get_script_and_batch_items, run_batch_items
from .config import init_logger, CONFIG, LOG, VERSION, arguments, MAGIC_SELECTION_INT
from .helpers import empty_queued_input_data, selector
from .parallel_runner import run_parallel_screens
from .progress_bar import setup_scroll_area, destroy_scroll_area


def check_for_original_automatix():
    p = subprocess.run('pip list | grep automatix', shell=True, stdout=subprocess.PIPE)

    for line in p.stdout.decode().split('\n'):
        if line.strip().split(' ', maxsplit=1)[0].strip() == 'automatix_cmd':
            raise Exception(
                'This package MUST NOT be installed along with the "automatix_cmd" package.'
                ' Both packages use the same entry point and module names and therefore'
                ' are conflicting. Please uninstall automatix AND automatix_cmd first,'
                ' THEN reinstall the package you want to use!')


def run_startup_script():
    if not CONFIG.get('startup_script'):
        return

    cmds = [os.path.expandvars(CONFIG["startup_script"])]
    cmds.extend(sys.argv)
    subprocess.run(cmds)


def setup(args: argparse.Namespace) -> float:
    """Setup logger and print version information"""
    init_logger(name=CONFIG['logger'], debug=args.debug)
    starttime = time()

    LOG.info(f'Automatix Version {VERSION}')
    LOG.info(f'Started at: {strftime("%a, %d %b %Y %H:%M:%S UTC", gmtime(starttime))}')

    configfile = CONFIG.get('config_file')
    if configfile:
        LOG.info(f'Using configuration from: {configfile}')
    else:
        LOG.warning('Configuration file not found or not configured. Using defaults.')

    return starttime


def check_screen():
    p = subprocess.run('screen -v', shell=True, stdout=subprocess.PIPE)
    screen_version = p.stdout.decode()
    if p.returncode != 0 or 'command not found' in screen_version:
        raise Exception('No GNU screen version found')
    if 'FAU' in screen_version:
        LOG.error(
            'Parallel processing only supported for the "GNU" version of screen. You have the "FAU" version.\n'
            'On MacOS you can try to install the GNU version via Homebrew: `brew install screen`.'
        )
        raise Exception('No supported GNU screen version found')


def main():
    check_for_original_automatix()

    if os.getenv('AUTOMATIX_SHELL'):
        print('You are running Automatix from an interactive shell of an already running Automatix!')
        empty_queued_input_data()
        answer = input('Do you really want to proceed? Then type "yes" and ENTER.\n')
        if answer != 'yes':
            sys.exit(0)

    args = arguments()
    run_startup_script()

    starttime = setup(args=args)

    script, batch_items = get_script_and_batch_items(args=args)

    if args.jump_to == MAGIC_SELECTION_INT:
        # next(iter(pi.items())) is here needed, because the pipeline items are all dictionaries with only one key.
        pipeline_items = [next(iter(pi.items())) for pi in script['pipeline']]
        args.jump_to = selector(
            entries=[
                (i, f'[{key}]: {cmd}')
                for i, (key, cmd) in enumerate(pipeline_items)
            ],
            message='Please choose index of desired start command:'
        )

    if args.vars_file and args.parallel:
        check_screen()
        run_parallel_screens(script=script, batch_items=batch_items, args=args)
        sys.exit(0)

    try:
        if CONFIG['progress_bar']:
            setup_scroll_area()

        run_batch_items(script=script, batch_items=batch_items, args=args)
    finally:
        if CONFIG['progress_bar']:
            destroy_scroll_area()

    if 'AUTOMATIX_TIME' in os.environ:
        LOG.info(f'The Automatix script took {round(time() - starttime)}s!')



import argparse
import pickle
import subprocess
from dataclasses import dataclass, field
from os import listdir, unlink
from os.path import isfile
from pathlib import Path
from time import sleep, strftime, gmtime

from .batch_runner import run_automatix_list
from .colors import yellow, green, red, cyan
from .config import LOG, init_logger, CONFIG
from .helpers import FileWithLock
from .progress_bar import setup_scroll_area, destroy_scroll_area

STATUS_TEMPLATE = 'waiting: {w}, running: {r}, user input required: {u}, finished: {f}'


@dataclass
class Autos:
    status_file: str
    time_id: int
    count: int
    tempdir: str

    max_parallel: int = 2
    waiting: list = field(default_factory=list)
    running: list = field(default_factory=list)
    user_input: list = field(default_factory=list)
    finished: list = field(default_factory=list)


def get_logfile_dir(time_id: int, scriptfile: str) -> str:
    human_readable_time = strftime('%Y-%m-%d_%H-%M-%S_UTC', gmtime(time_id))
    return f'{CONFIG.get("logfile_dir")}/{human_readable_time}__{Path(scriptfile).stem}'


def get_files(tempdir: str) -> set:
    return {f for f in listdir(tempdir) if isfile(f'{tempdir}/{f}') and f.startswith('auto')}


def print_status(autos: Autos):
    print(STATUS_TEMPLATE.format(
        w=yellow(len(autos.waiting)),
        r=cyan(len(autos.running)),
        u=red(len(autos.user_input)),
        f=green(f'{len(autos.finished)}/{autos.count}'),
    ))


def get_screen_status_line(label: str) -> str:
    status_line = yellow(f'### {label}')
    status_line += f' | detach: {cyan("<ctrl>+a d")}'
    status_line += f' | copy mode: {cyan("<ctrl>+a Esc")}'
    status_line += f' | abort copy mode: {cyan("Esc ")}'
    # It seems we need the space after this Esc,  ^^^
    # otherwise the color reset escape sequence stops working.
    return status_line


def check_for_status_change(autos: Autos, status_file: str):
    with FileWithLock(status_file, 'r+') as sf:
        for line in sf:
            if not line:
                continue
            LOG.debug(f'Line: {line}')
            auto_file, status = line.strip().split(':')
            LOG.debug(f'Got {auto_file}:{status}')
            match status:
                case 'max_parallel':
                    # In this case we misuse the "auto_file" part as number
                    # for how many parallel screens are allowed.
                    autos.max_parallel = int(auto_file)
                    LOG.info(f'Now process max {auto_file} screens parallel')
                case 'user_input_remove':
                    autos.user_input.remove(auto_file)
                case 'user_input_add':
                    autos.user_input.append(auto_file)
                    LOG.info(f'{auto_file} is waiting for user input')
                case 'finished':
                    autos.running.remove(auto_file)
                    autos.finished.append(auto_file)
                    LOG.info(f'{auto_file} finished')
                case _:
                    LOG.warning(f'[{auto_file}] Unrecognized status "{status}"\n')
        # Empty status file as all status have been processed
        sf.truncate(0)


def run_manage_loop(tempdir: str, time_id: int):
    status_file = f'{tempdir}/{time_id}_overview'
    auto_files = sorted(get_files(tempdir))
    autos = Autos(status_file=status_file, time_id=time_id, count=len(auto_files), waiting=auto_files, tempdir=tempdir)
    with open(f'{tempdir}/{next(iter(auto_files))}', 'rb') as f:
        logfile_dir = pickle.load(f)['logfile_dir']

    LOG.info(f'Found {autos.count} files to process. Screens name are like "{time_id}_autoX"')
    LOG.info('To switch screens detach from this screen via "<ctrl>+a d".')
    LOG.info('To scroll back in history press "<ctrl>+a Esc" to enable "copy mode". Switch back with "Esc".')
    LOG.info('You can modify this behaviour by screen configuration options (`~/.screenrc`).')

    open(status_file, 'a').close()
    try:
        while len(autos.finished) < autos.count:
            if len(autos.running) < autos.max_parallel and autos.waiting:
                auto_file = autos.waiting.pop(0)
                autos.running.append(auto_file)

                auto_path = f'{tempdir}/{auto_file}'
                with open(auto_path, 'rb') as f:
                    auto_file_data = pickle.load(file=f)
                status_line = get_screen_status_line(label=auto_file_data['label'])

                session_name = f'{time_id}_{auto_file}'
                logfile_path = f'{logfile_dir}/{auto_file}.log'

                LOG.info(f'Starting new screen at {session_name}')
                subprocess.run([
                    'screen', '-d', '-m', '-S', session_name,
                    '-h', '100000',
                    '-L', '-Logfile', logfile_path,
                    'automatix-from-file', tempdir, str(time_id), auto_file
                ])
                subprocess.run(['screen', '-S', session_name, '-X', 'hardstatus', 'alwayslastline'])
                subprocess.run(['screen', '-S', session_name, '-X', 'hardstatus', 'string', status_line])

            check_for_status_change(autos=autos, status_file=status_file)

            print_status(autos=autos)

            # Write status to file for screen switch loop
            with FileWithLock(f'{tempdir}/autos', 'wb') as f:
                pickle.dump(obj=autos, file=f)

            sleep(1)

        LOG.info(f'All parallel screen reported finished ({len(autos.finished)}/{autos.count}).')
    except Exception as exc:
        LOG.exception(exc)
        sleep(60)  # For debugging
    finally:
        with open(f'{tempdir}/{time_id}_finished', 'w') as fifo:
            fifo.write('finished')


def run_manager():
    parser = argparse.ArgumentParser(
        description='Automatix manager for parallel processing (called by the automatix main programm)',
        epilog='Explanations and README at https://github.com/vanadinit/automatix',
    )
    parser.add_argument('tempdir')
    parser.add_argument('time_id')
    parser.add_argument(
        '--debug', '-d',
        action='store_true',
        help='activate debug log level',
    )
    args = parser.parse_args()

    init_logger(name=CONFIG['logger'], debug=args.debug)
    run_manage_loop(tempdir=args.tempdir, time_id=int(args.time_id))


def run_auto(tempdir: str, time_id: int, auto_file: str):
    auto_path = f'{tempdir}/{auto_file}'

    def send_status(status: str):
        with FileWithLock(f'{tempdir}/{time_id}_overview', 'a') as sf:
            sf.write(f'{auto_file}:{status}\n')

    with open(auto_path, 'rb') as f:
        auto_file_data = pickle.load(file=f)
    try:
        run_automatix_list(automatix_list=auto_file_data['autolist'], send_status_callback=send_status)
    finally:
        send_status('finished')
        unlink(auto_path)


def run_auto_from_file():
    parser = argparse.ArgumentParser(
        description='Automatix from file for parallel processing (called by the automatix-manager)',
        epilog='Explanations and README at https://github.com/vanadinit/automatix',
    )
    parser.add_argument('tempdir')
    parser.add_argument('time_id')
    parser.add_argument('auto_file')
    args = parser.parse_args()

    try:
        if CONFIG['progress_bar']:
            setup_scroll_area()

        run_auto(tempdir=args.tempdir, time_id=args.time_id, auto_file=args.auto_file)
    finally:
        if CONFIG['progress_bar']:
            destroy_scroll_area()



import os
import pickle
import subprocess
from argparse import Namespace
from collections import defaultdict
from tempfile import TemporaryDirectory
from time import time

from .batch_runner import create_automatix_list
from .config import LOG
from .parallel import get_logfile_dir, get_screen_status_line
from .parallel_ui import screen_switch_loop


def get_batch_groups(batch_items: list) -> dict:
    batch_groups = defaultdict(list)
    for batch_item in batch_items:
        if group := batch_item.get('group'):
            assert group != '_default_', 'The group name "_default_" is reserved. Please use something different.'
            batch_groups[group].append(batch_item)
        else:
            batch_groups['_default_'].append(batch_item)
    return batch_groups


def write_auto_file(auto_id: str, label: str, autolist: list, tempdir: str, logfile_dir: str):
    with open(f'{tempdir}/auto{auto_id}', 'wb') as f:
        pickle.dump(obj={
            'autolist': autolist,
            'auto_file': f'{tempdir}/auto{auto_id}',
            'label': label,
            'logfile_dir': logfile_dir,
        }, file=f)


def create_auto_files(script: dict, batch_items: list, args: Namespace, tempdir: str, logfile_dir: str):
    LOG.info(f'Using temporary directory to save object files: {tempdir}')

    batch_groups = get_batch_groups(batch_items=batch_items)
    default_group = batch_groups.pop('_default_', [])

    digits = len(str(len(batch_groups) + len(default_group)))

    i = 1
    # Create grouped automatix files for each "real" group
    for i, (group, items) in enumerate(batch_groups.items(), start=1):
        write_auto_file(
            auto_id=str(i).rjust(digits, '0'),
            autolist=create_automatix_list(script=script, batch_items=items, args=args),
            label=f'Group: {group}',
            tempdir=tempdir,
            logfile_dir=logfile_dir,
        )

    # All rows/batch_items in the default group get their own automatix file and screen
    for j, batch_item in enumerate(default_group, start=i + 1):
        auto_id = str(j).rjust(digits, '0')

        # Get label before create_automatix_list, where the label field is removed
        label = batch_item.get('label', f'auto{auto_id}')

        autolist = create_automatix_list(script=script, batch_items=[batch_item], args=args)
        write_auto_file(
            auto_id=auto_id,
            autolist=autolist,
            label=label,
            tempdir=tempdir,
            logfile_dir=logfile_dir,
        )


def display_screen_control_hints():
    LOG.notice('--- Please read and understand these hints for controlling screen sessions before you proceed ---')
    LOG.notice('- If you switched to a screen session you can detach the session with "<ctrl>+a d".'
               ' This takes you back to the information interface.')
    LOG.notice('- To scroll back in history press "<ctrl>+a Esc" to enable "copy mode". Switch back with "Esc".')
    LOG.notice('- You can modify this behaviour by screen configuration options (`~/.screenrc`).')
    LOG.notice('- To suppress this message set the `AUTOMATIX_SUPPRESS_SCREEN_CONTROL_NOTICE` environment variable.')
    input('Press ENTER to continue...')


def run_parallel_screens(script: dict, batch_items: list, args: Namespace):
    LOG.info('Preparing automatix objects for parallel processing')

    with TemporaryDirectory() as tempdir:
        time_id = round(time())
        logfile_dir = get_logfile_dir(time_id=time_id, scriptfile=args.scriptfile)
        os.makedirs(logfile_dir)
        os.mkfifo(f'{tempdir}/{time_id}_finished')

        create_auto_files(script=script, batch_items=batch_items, args=args, tempdir=tempdir, logfile_dir=logfile_dir)

        LOG.info(f'Created directory for logfiles at {logfile_dir}')

        cmds = [
            'screen', '-d', '-m', '-S', f'{time_id}_overview',
            '-h', '100000',
            '-L', '-Logfile', f'{logfile_dir}/overview.log',
            'automatix-manager', tempdir, str(time_id),
        ]
        if args.debug:
            cmds.append('--debug')

        subprocess.run(cmds)

        status_line = get_screen_status_line(label='Manager screen')
        subprocess.run(['screen', '-S', f'{time_id}_overview', '-X', 'hardstatus', 'alwayslastline'])
        subprocess.run(['screen', '-S', f'{time_id}_overview', '-X', 'hardstatus', 'string', status_line])

        LOG.info(f'Overview / manager screen started at "{time_id}_overview".')

        LOG.info('Start loop with information to switch between running screens.\n')
        if not os.getenv('AUTOMATIX_SUPPRESS_SCREEN_CONTROL_NOTICE'):
            display_screen_control_hints()

        screen_switch_loop(tempdir=tempdir)

        with open(f'{tempdir}/{time_id}_finished') as fifo:
            print()
            LOG.info('Wait for overview to finish')
            for _ in fifo:
                LOG.info('Automatix finished parallel processing')
    LOG.info('Temporary directory cleaned up')
    LOG.info(f'All logfiles are available at {logfile_dir}')



import curses
import pickle
import subprocess
from os.path import isfile
from textwrap import wrap
from time import sleep

from .config import LOG
from .helpers import FileWithLock
from .parallel import Autos


class CursesWriter:
    def __init__(self, stdscr: curses.window):
        self.stdscr = stdscr

        # Terminal setup
        curses.curs_set(0)  # Hide cursor
        stdscr.nodelay(True)  # Make getch() non-blocking

        # Initialize colors
        curses.start_color()
        curses.use_default_colors()
        curses.init_pair(1, curses.COLOR_YELLOW, -1)  # -1 for default background
        curses.init_pair(2, curses.COLOR_CYAN, -1)
        curses.init_pair(3, curses.COLOR_RED, -1)
        curses.init_pair(4, curses.COLOR_GREEN, -1)

        self.yellow = curses.color_pair(1)
        self.cyan = curses.color_pair(2)
        self.red = curses.color_pair(3)
        self.green = curses.color_pair(4)

        self.h, self.w = stdscr.getmaxyx()

        self.input_buffer = ''
        self.current_line = 0

    def add_text(self, text: str, start: int = 0, attr: int | None = None, append_line: bool = False):
        if append_line:
            self.current_line, x = self.stdscr.getyx()
            start = start if start else x + 1

        available_width = self.w - start - 1
        if available_width <= 0:  # Not enough space to print anything
            # Move to the next line anyway to avoid infinite loops on tiny screens
            self.add_empty_line()
            return

        wrapped_lines = wrap(text, width=available_width)

        # If the original text was empty, textwrap returns an empty list.
        # Ensure we "print" an empty line to advance the cursor.
        if not wrapped_lines:
            wrapped_lines.append('')

        for line in wrapped_lines:
            if attr is not None:
                self.stdscr.addstr(self.current_line, start, line, attr)
            else:
                self.stdscr.addstr(self.current_line, start, line)
            self.current_line += 1

    def add_empty_line(self):
        self.current_line += 1
        self.stdscr.move(self.current_line, 0)

    def clear(self):
        self.stdscr.clear()
        self.current_line = 0


def handle_exit(exc: Exception) -> str:
    if isinstance(exc, KeyboardInterrupt):
        LOG.error('You decided to quit the UI, but screens are most likely still running. Please decide what to do:')
    else:
        LOG.exception(exc)
        print()
        LOG.error('An unexpected error occurred. Please decide what to do:')
    match input(
        'Press "q" and Enter to quit.'
        ' This will cause this programm to terminate.'
        ' Check "screen -list" afterwards for still running screens.\n'
        ' Press something else to restart the UI, where you can switch to the different screens.\n'
    ):
        case 'q':
            raise SystemExit(130)
        case _:
            return 'restart'


def draw_status(cw: CursesWriter, autos: Autos):
    cw.clear()

    cw.add_text(f'Automatix Status (max parallel: {autos.max_parallel})')
    cw.add_text('-' * (cw.w - 2))

    cw.add_text('waiting: ')
    cw.add_text(str(len(autos.waiting)), attr=cw.yellow, append_line=True)
    cw.add_text(', running: ', append_line=True)
    cw.add_text(str(len(autos.running)), attr=cw.cyan, append_line=True)
    cw.add_text(', user input required: ', append_line=True)
    cw.add_text(str(len(autos.user_input)), attr=cw.red, append_line=True)
    cw.add_text(', finished: ', append_line=True)
    cw.add_text(f'{len(autos.finished)}/{autos.count}', attr=cw.green, append_line=True)

    cw.add_text('-' * (cw.w - 2))
    cw.add_empty_line()

    cw.add_text('Waiting:             ', start=2)
    cw.add_text(', '.join(autos.waiting), start=22, append_line=True)
    cw.add_text('Running:             ', start=2)
    cw.add_text(', '.join(autos.running), start=22, append_line=True)
    cw.add_text('User input needed: ', attr=cw.red, start=2)
    cw.add_text(', '.join(autos.user_input), attr=cw.red, start=22, append_line=True)
    cw.add_text('Finished:            ', start=2)
    cw.add_text(', '.join(autos.finished), start=22, append_line=True)

    cw.current_line = cw.h - 6
    cw.add_text(f'Working directory: {autos.tempdir}')
    cw.add_text("-" * (cw.w - 1))
    cw.add_text(
        'Options: [o] Overview | [n] Next Input | [X] to autoX | [mX] max parallel to X | [q] Quit',
        start=2,
    )
    cw.add_text(f'Input: {cw.input_buffer}', start=2)

    cw.stdscr.refresh()


def handle_answer(answer: str, autos: Autos) -> str | None:
    if answer == 'q':
        raise KeyboardInterrupt('Quit UI by user')
    elif answer == 'o':
        return f'{autos.time_id}_overview'
    elif answer == 'n' and autos.user_input:
        return f'{autos.time_id}_{next(iter(autos.user_input))}'
    elif answer.startswith('m'):
        try:
            max_parallel = int(answer[1:])
            with FileWithLock(autos.status_file, 'a') as sf:
                sf.write(f'{max_parallel}:max_parallel\n')
        except (ValueError, IndexError):
            pass  # Ignore invalid input
    else:
        try:
            number = int(answer)
            return f'{autos.time_id}_auto{str(number).rjust(len(str(autos.count)), "0")}'
        except ValueError:
            pass  # Ignore invalid input


def process_user_input(cw: CursesWriter, autos: Autos) -> str | None:
    key = cw.stdscr.getch()
    if key == -1:  # No input
        return None

    char = chr(key)
    if key in [curses.KEY_ENTER, 10, 13]:
        answer = cw.input_buffer.lower()
        cw.input_buffer = ''

        if new_screen := handle_answer(answer=answer, autos=autos):
            return new_screen
    elif key in [curses.KEY_BACKSPACE, 127]:
        cw.input_buffer = cw.input_buffer[:-1]
    elif char.isalnum():
        cw.input_buffer += char


def parallel_ui(stdscr: curses.window, tempdir: str) -> tuple[str, str | None]:
    cw = CursesWriter(stdscr=stdscr)

    # Wait until the status file exists
    while not isfile(f'{tempdir}/autos'):
        stdscr.clear()
        stdscr.addstr(0, 0, 'Waiting for the manager process to start...')
        stdscr.refresh()
        sleep(0.5)

    while True:
        # 1. Load data
        with FileWithLock(f'{tempdir}/autos', 'rb') as f:
            autos = pickle.load(file=f)

        # 2. Draw status
        draw_status(cw=cw, autos=autos)

        # 3. Check if all processes have finished
        if len(autos.running) + len(autos.waiting) + len(autos.user_input) == 0:
            cw.stdscr.addstr(cw.h - 1, 2, 'All processes have finished. Press "q" to exit.', cw.green)
            cw.stdscr.refresh()
            # Wait until the user presses 'q'
            cw.stdscr.nodelay(False)
            while cw.stdscr.getch() not in [ord('q'), ord('Q')]:
                sleep(0.1)
            return 'quit', None

        # 4. Process user input
        screen_to_switch = process_user_input(cw=cw, autos=autos)
        if screen_to_switch:
            return 'restart', screen_to_switch

        sleep(0.2)  # Short pause to reduce CPU load


def screen_switch_loop(tempdir: str):
    while True:
        try:
            exit_reason, screen_to_switch = curses.wrapper(parallel_ui, tempdir)
            if screen_to_switch:
                print(f"Switching to screen '{screen_to_switch}'... (Return with <ctrl>+a d)")
                subprocess.run(['screen', '-r', screen_to_switch])
        except curses.error as exc:
            LOG.error(f"Curses error: {exc}")
            LOG.error("Could not start the curses interface. Is the terminal compatible?")
            exit_reason = handle_exit(exc=exc)
        except (KeyboardInterrupt, Exception) as exc:
            exit_reason = handle_exit(exc=exc)

        if exit_reason != 'restart':
            break



from unittest import TestCase

from automatix.environment import AttributedDict, AttributedDummyDict

tc = TestCase()


def test__attributed_dict():
    test_dict = AttributedDict()

    test_dict['a'] = 5
    assert test_dict.a == 5

    test_dict.b = 'abcde'
    assert test_dict['b'] == 'abcde'

    with tc.assertRaises(AttributeError):
        _ = test_dict.not_existent

    with tc.assertRaises(KeyError):
        _ = test_dict['not_existent']


def test__attributed_dummy_dict():
    test_dict = AttributedDummyDict('DICTNAME')

    test_dict['a'] = 5
    assert test_dict.a == 5

    test_dict.b = 'abcde'
    assert test_dict['b'] == 'abcde'

    assert test_dict.not_existent == test_dict['not_existent'] == '{DICTNAME.not_existent}'
    assert test_dict.__name__ == 'DICTNAME'



import os
from sys import stdin
from time import sleep
from typing import List

import yaml
from termios import tcflush, TCIFLUSH

yaml.warnings({'YAMLLoadWarning': False})


def read_yaml(yamlfile: str) -> dict:
    with open(yamlfile) as file:
        return yaml.load(file.read(), Loader=yaml.SafeLoader)


def empty_queued_input_data():
    tcflush(stdin, TCIFLUSH)


def selector(entries: List[tuple], message: str = 'Found multiple entries, please choose:'):
    """Provides a command line interface for selecting from multiple entries
    :param entries: List of Tuples(entry: Any, label: str)
    :param message: message to be displayed for selection
    """
    match len(entries):
        case 0:
            return None
        case 1:
            return entries[0][0]
        case _:
            entry_labels = '\n'.join([
                f'{index}: {entry[1]}'
                for index, entry in enumerate(entries)
            ])
            try:
                empty_queued_input_data()
                return entries[int(input(f'\n{message} \n{entry_labels}\n\nYour choice:\n'))][0]
            except (ValueError, IndexError):
                print('Invalid answer! Please try again and type the number of your desired answer.')
                return selector(entries)


class FileWithLock:
    def __init__(self, file_path: str, method: str):
        self.file_path = file_path
        self.method = method
        self.file_obj = None

    def __enter__(self):
        get_lock(self.file_path)
        self.file_obj = open(self.file_path, self.method)
        return self.file_obj

    def __exit__(self, type, value, traceback):
        self.file_obj.close()
        release_lock(self.file_path)


def get_lock(file_path: str):
    while True:
        try:
            os.mkdir(f'{file_path}.lock')
        except FileExistsError:
            sleep(1)
            continue
        break


def release_lock(file_path: str):
    os.rmdir(f'{file_path}.lock')



import sys
from argparse import Namespace
from copy import deepcopy
from csv import DictReader
from typing import Callable

from .automatix import Automatix
from .command import SkipBatchItemException, AbortException
from .config import CONFIG, get_script, LOG, update_script_from_row, collect_vars, SCRIPT_FIELDS


def get_script_and_batch_items(args: Namespace) -> (dict, list):
    script = get_script(args=args)

    # Empty item means: there is nothing to update, take the script as it is
    batch_items: list[dict] = [{}]
    if args.vars_file:
        with open(args.vars_file) as csvfile:
            batch_items = list(DictReader(csvfile))

    return script, batch_items


def create_automatix_list(script: dict, batch_items: list, args: Namespace) -> list[Automatix]:
    automatix_list = []
    for i, row in enumerate(batch_items, start=1):
        script_copy = deepcopy(script)
        script_copy['_batch_mode'] = len(batch_items) > 1
        script_copy['_batch_items_count'] = len(batch_items)

        update_script_from_row(row=row, script=script_copy, index=i)

        variables = collect_vars(script_copy)

        auto = Automatix(
            script=script_copy,
            variables=variables,
            config=CONFIG,
            script_fields=SCRIPT_FIELDS,
            cmd_args=args,
            batch_index=i,
        )
        automatix_list.append(auto)
    return automatix_list


def run_automatix_list(automatix_list: list[Automatix], send_status_callback: Callable = None):
    for auto in automatix_list:
        auto.set_command_count()
        auto.env.attach_logger()
        auto.env.reinit_logger()
        if send_status_callback:
            auto.env.send_status = send_status_callback
        try:
            auto.run()
        except SkipBatchItemException as exc:
            LOG.info(str(exc))
            LOG.notice('=====> Jumping to next batch item.')
            continue
        except AbortException as exc:
            sys.exit(int(exc))
        except KeyboardInterrupt:
            print()
            LOG.warning('Aborted by user. Exiting.')
            sys.exit(130)


def run_batch_items(script: dict, batch_items: list, args: Namespace):
    automatix_list = create_automatix_list(script=script, batch_items=batch_items, args=args)
    run_automatix_list(automatix_list=automatix_list)
